<?php

namespace Drupal\bokbasen_checkout\ApiClient;

use Drupal\bokbasen_checkout\Exception\BokbasenException;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\State\StateInterface;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\ClientException;
use GuzzleHttp\RequestOptions;
use Psr\Log\LoggerAwareInterface;
use Psr\Log\LoggerAwareTrait;

/**
 * Authentication class to use against Bokbasen's web services
 *
 * @link https://bokbasen.jira.com/wiki/spaces/api/pages/2994962433/Authentication+Service
 */
class Authenticator implements LoggerAwareInterface {

  use LoggerAwareTrait;

  /**
   * Http client.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $httpClient;

  /**
   * State service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $stateService;

  /**
   * Constructs Authenticator.
   */
  public function __construct(ClientInterface $http_client, StateInterface $stateService, LoggerChannelFactoryInterface $logger) {
    $this->httpClient = $http_client;
    $this->stateService = $stateService;
    $this->logger = $logger->get('bokbasen_checkout');
  }

  /**
   * Get valid access token or create new one.
   */
  public function getToken(string $clientId, string $clientSecret, string $audience, string $url) {
    $token = $this->stateService->get("bokbasen_access_token.{$audience}", []);
    if (!isset($token['access_token']) || $token['expire'] <= $_SERVER['REQUEST_TIME']) {
      $token = $this->authenticate($clientId, $clientSecret, $audience, $url);
      $this->stateService->set("bokbasen_access_token.{$audience}", $token);
      $this->logger->debug('Created new access_token for Bokbasen API');
    }
    if (!isset($token['access_token'])) {
      throw new BokbasenException('Unable to authorize in Bokbasen API');
    }
    return $token;
  }

  /**
   * Authenticate user and return token.
   */
  public function authenticate(string $clientId, string $clientSecret, string $audience, string $url) {
    $body = [
      'client_id' => $clientId,
      'client_secret' => $clientSecret,
      'audience' => $audience,
      'grant_type' => 'client_credentials'
    ];
    try {
      $response = $this->httpClient->request('POST', $url, [
        RequestOptions::HEADERS => ['content-type' => 'application/json'],
        RequestOptions::ALLOW_REDIRECTS => [
          'strict' => TRUE,
        ],
        RequestOptions::BODY => json_encode($body),
      ]);
      $decodedResponse = json_decode((string) $response->getBody(), TRUE);
      return [
        'access_token' => $decodedResponse['access_token'],
        'expire' => $_SERVER['REQUEST_TIME'] + $decodedResponse['expires_in'],
      ];
    } catch (ClientException $exception) {
      if ($exception->getCode() === 401) {
        $this->logger->warning('The authentication attempt in Bokbasen API resulted in a `401 Unauthorized`');
        return [];
      }
      else {
        throw $exception;
      }
    }
  }
}