<?php

namespace Drupal\bokbasen_checkout\ApiClient;

use Drupal\bokbasen_checkout\Exception\BokbasenException;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Site\Settings;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\StringTranslation\TranslationManager;
use Drupal\serialization\Encoder\JsonEncoder;
use Drupal\serialization\Encoder\XmlEncoder;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\RequestOptions;
use Symfony\Component\Serializer\Normalizer\GetSetMethodNormalizer;
use Symfony\Component\Serializer\Serializer;

/**
 * Class BokbasenApiClientFactory.
 */
abstract class BokbasenApiClientBase implements BokbasenApiClientInterface {

  use StringTranslationTrait;

  /**
   * {@inheritdoc}
   */
  abstract public function getBaseUrl();

  /**
   * {@inheritdoc}
   */
  abstract public function getAudience();

  /**
   * {@inheritdoc}
   */
  abstract public function getCredentials();

  /**
   * Logger.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger = NULL;


  /**
   * Json Encoder.
   *
   * @var \Drupal\serialization\Encoder\JsonEncoder
   */
  protected $jsonEncoder;

  /**
   * XML encoder.
   *
   * @var \Drupal\serialization\Encoder\XmlEncoder
   */
  protected $xmlEncoder;

  /**
   * Module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * Module handler.
   *
   * @var \Drupal\bokbasen_checkout\ApiClient\Authenticator
   */
  protected $authenticator;

  /**
   * Http client.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $httpClient;

  /**
   * Constructs a BokbasenApiClient object.
   */
  public function __construct(JsonEncoder $json_encoder, XmlEncoder $xml_encoder, LoggerChannelFactoryInterface $logger, ModuleHandlerInterface $module_handler, TranslationManager $string_translation, ClientInterface $http_client, Authenticator $authenticator) {
    $this->jsonEncoder = $json_encoder;
    $this->xmlEncoder = $xml_encoder;
    $this->stringTranslation = $string_translation;
    $this->moduleHandler = $module_handler;
    $this->logger = $logger->get('bokbasen_checkout');
    $serializer = new Serializer([new GetSetMethodNormalizer()]);
    $this->xmlEncoder->setSerializer($serializer);
    $this->httpClient = $http_client;
    $this->authenticator = $authenticator;
  }

  /**
   * {@inheritdoc}
   */
  public function getRequest(string $path, array $headers = [], string $encode_format = JsonEncoder::FORMAT) {
    $credentials = $this->getCredentials();
    $token = $this->authenticator->getToken($credentials['client_id'], $credentials['client_secret'], $this->getAudience(), $this->getLoginUrl());
    $headers['Authorization'] = "Bearer {$token['access_token']}";

    $results = [];
    if (!isset($headers['Accept']) && $encode_format === JsonEncoder::FORMAT) {
      $headers['Accept'] = 'application/json';
    }

    $response = $this->httpClient->request('GET', $this->getBaseUrl() . $path, [
      RequestOptions::HEADERS => $headers,
      RequestOptions::ALLOW_REDIRECTS => [
        'strict' => TRUE,
      ],
    ]);

    if ($response->getStatusCode() == 200) {
      $results = $response->getBody()->getContents();
    }
    // Decode json.
    if ($results && $encode_format === JsonEncoder::FORMAT) {
      return $this->jsonEncoder->decode($results, JsonEncoder::FORMAT);
    }
    // Decode XML.
    if ($results && $encode_format === 'xml') {
      return $this->xmlEncoder->decode($results, $encode_format);
    }

    if ($response->getStatusCode() == 302) {
      if ($response->hasHeader('Location')) {
        return $response->getHeader('Location');
      }
    }
    else {
      $this->logger->notice($response->getBody());
    }

    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function jsonPostRequest(string $path, array $body, array $headers = [], string $encode_format = JsonEncoder::FORMAT) {
    $credentials = $this->getCredentials();
    $token = $this->authenticator->getToken($credentials['client_id'], $credentials['client_secret'], $this->getAudience(), $this->getLoginUrl());
    $headers['Authorization'] = "Bearer {$token['access_token']}";

    if (!isset($headers['Accept']) && $encode_format === JsonEncoder::FORMAT) {
      $headers['Accept'] = 'application/json';
    }

    $body = json_encode($body);

    if ($body === false) {
      throw new BokbasenException('Unable to convert data to json');
    }

    $headers = array_merge(
      ['Content-Type' => 'application/json'],
      $headers
    );

    $response = $this->httpClient->request('POST', $this->getBaseUrl() . $path, [
      RequestOptions::HEADERS => $headers,
      RequestOptions::ALLOW_REDIRECTS => [
        'strict' => TRUE,
      ],
      RequestOptions::BODY => $body,
    ]);

    $results = $response->getBody();
    if ($response->getStatusCode() == 200 && $encode_format === JsonEncoder::FORMAT) {
      return $this->jsonEncoder->decode($results, JsonEncoder::FORMAT);
    }
    if ($response->getStatusCode() == 201) {
      if ($response->hasHeader('Location')) {
        return $response->getHeader('Location');
      }

      return TRUE;
    }
    else {
      $this->logger->notice($results->getContents());
    }

    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function getLoginUrl() {
    $is_test_mode = $this->isTestMode();

    return $is_test_mode ? 'https://auth.stage.bokbasen.io/oauth/token' : 'https://auth.bokbasen.io/oauth/token';
  }

  /**
   * {@inheritdoc}
   */
  public function isTestMode() {
    return Settings::get('bokbasen_test_mode', FALSE);
  }

}
