<?php

namespace Drupal\bokbasen_checkout\ApiClient;

use Drupal\bokbasen_checkout\Exception\BokbasenException;
use Drupal\Core\Site\Settings;

/**
 * Bokbasen Distribution API implementation.
 */
class BokbasenDistribution extends BokbasenApiClientBase {

  const API_VERSION = 'v2';
  const CONTENT_STATUS_READY = 'READY';
  const CONTENT_STATUS_FAILED = 'FAILED';
  const CONTENT_PATH = 'content';

  /**
   * {@inheritdoc}
   */
  public function getBaseUrl() {
    return Settings::get('bokbasen_dds_api_url', 'https://api.bokbasen.io');
  }

  /**
   * {@inheritdoc}
   */
  public function getAudience() {
    return Settings::get('bokbasen_dds_audience', 'https://api.bokbasen.io/dds/');
  }

  /**
   * {@inheritdoc}
   */
  public function getCredentials() {
    return Settings::get('bokbasen_dds_credentials', [
      'client_id' => '',
      'client_secret' => '',
    ]);
  }

  /**
   * Get inventory.
   *
   * Next or after should be provided when no isbn.
   * @link https://bokbasen.jira.com/wiki/spaces/api/pages/61964312/Inventory
   */
  public function getInventory(string $isbn = NULL, string $after = NULL, string $next = NULL) {
    $query = '';
    if (!$isbn) {
      $query = sprintf('after=%s&next=%s', $after, $next);
    }
    $results = $this->getRequest(sprintf('/dds/%s/%s/%s?%s', 'inventory', self::API_VERSION, $isbn, $query), [], 'xml');
    if (!empty($results['entry'])) {
      if (isset($results['entry']['id'])) {
        $results['entry'] = [
          $results['entry'],
        ];
      }

      return $results['entry'];
    }

    return [];
  }

  /**
   * Place order.
   *
   * @link https://bokbasen.jira.com/wiki/spaces/api/pages/61964322/Orders+and+reporting
   *
   * @param array $order_data
   *   User data.
   *   array(
   *      'orderid' => '1111' - store internal Id*
   *      'resid' => '2222' - product DDS ID*
   *      'spd' => '21' - price*
   *      'id' => 'dsd3-324ds-d' - DDS ID foe end consumer
   *      'firstname' => 'example name' - first name of person who buying
   *      'lastname' => 'example_name' - last name of person who buying
   *      'excerpt' => 1 - is for excerpt instead of actual book(defaults to 0)
   *      'email' => optional, for internal usage
   *   )
   *
   * @return mixed
   *   Content id or FALSE.
   */
  public function placeOrder(array $order_data) {
    $headers['Content-Type'] = 'application/json';
    $this->moduleHandler->alter('bokbasen_order', $order_data);
    $results = $this->jsonPostRequest(sprintf('/dds/%s/v1', 'order'), $order_data, $headers);
    if (is_array($results)) {
      return reset($results);
    }
    // User cant buy this product.
    $this->logger->notice(
      $this->t(
        'User with email: %email cant buy product: %product_id within order: %order',
        [
          '%email' => $order_data['email'] ?: 'not set',
          '%product_id' => $order_data['resid'],
          '%order' => $order_data['orderid'],
        ]
      )
    );

    return $results;
  }

  /**
   * Get download link for the content.
   *
   * @link https://bokbasen.jira.com/wiki/spaces/api/pages/62291987/Content+download#Get-content-(based-on-fulfillment-ID)
   */
  public function getDownloadContentLink(string $fulfilment_id) {
    $results = $this->getRequest(sprintf('/dds/content/v1/%s', $fulfilment_id));
    $download_link = NULL;
    if ($results) {
      $status = $results['status'];
      while ($status !== self::CONTENT_STATUS_READY) {
        $results = $this->getRequest(sprintf('/dds/content/v1/%s', $fulfilment_id));
        $status = $results['status'];
        // Failed.
        if ($status === self::CONTENT_STATUS_FAILED) {
          break;
        }
        sleep(1);
      }
      if ($status === self::CONTENT_STATUS_READY) {
        $download_link = $results['downloadURL'];
      }
      else {
        throw new BokbasenException($this->t('Bokbasen can not fetch a download link for the fulfilmentid: %id.', ['%id' => $fulfilment_id]));
      }
    }

    return $download_link;
  }

}
